/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "velox/dwio/common/UnpackBase.h"

namespace facebook::velox::dwio::common {

template <int8_t i>
auto as8x64(__m512i x) {
  return _mm512_cvtepu32_epi64(_mm512_extracti64x4_epi64(x, i));
}

auto as16x16(__m512i x) {
  return _mm512_cvtepi32_epi16(x);
}

auto as16x8(__m512i x) {
  return _mm512_cvtepi32_epi8(x);
}

template <typename T>
void store16Ints(__m512i sixteenInts, int32_t i, T* result) {
  if (sizeof(T) == 4) {
    _mm512_storeu_si512(result + i, sixteenInts);
  } else if (sizeof(T) == 2) {
    _mm256_storeu_si256(
        reinterpret_cast<__m256i*>(result + i), as16x16(sixteenInts));
  } else if (sizeof(T) == 1) {
    _mm_store_si128(
        reinterpret_cast<__m128i*>(result + i), as16x8(sixteenInts));
  } else {
    _mm512_storeu_si512(result + i, as8x64<0>(sixteenInts));
    _mm512_storeu_si512(result + i + 8, as8x64<1>(sixteenInts));
  }
}

template <typename T>
inline T* addBytes(T* pointer, int32_t bytes) {
  return reinterpret_cast<T*>(reinterpret_cast<uint64_t>(pointer) + bytes);
}

typedef int32_t __m512si __attribute__((__vector_size__(64), __may_alias__));

typedef int32_t __m512si_u
    __attribute__((__vector_size__(64), __may_alias__, __aligned__(1)));

template <typename T>
inline __m512i as512i(T x) {
  return reinterpret_cast<__m512i>(x);
}

template <typename T>
inline __m512si as16x32(T x) {
  return reinterpret_cast<__m512si>(x);
}

template <typename T>
FOLLY_ALWAYS_INLINE __m512i gather16Sparse(
    const uint64_t* bits,
    int32_t bitOffset,
    const int32_t* rows,
    uint8_t width,
    int32_t i,
    T* result) {
  uint64_t kMask = bits::lowMask(width);
  const auto masks = as16x32(_mm512_set1_epi32(kMask));
  constexpr __m512si kMultipliers = {
      256, 128, 64, 32, 16, 8, 4, 2, 256, 128, 64, 32, 16, 8, 4, 2};

  auto indices =
      *reinterpret_cast<const __m512si_u*>(rows + i) * width + bitOffset;
  __m512si multipliers;
  if (width % 8 != 0) {
    multipliers = (__m512si)_mm512_permutexvar_epi32(
        as512i(indices & 7), as512i(kMultipliers));
  }
  auto byteIndices = indices >> 3;
  auto data = reinterpret_cast<__m512si>(_mm512_i32gather_epi32(
      as512i(byteIndices), reinterpret_cast<const int*>(bits), 1));
  if (width % 8 != 0) {
    data = (data * multipliers) >> 8;
  }
  auto sixteenInts = as512i(data & masks);
  store16Ints(sixteenInts, i, result);
  return sixteenInts;
}

} // namespace facebook::velox::dwio::common